@@ -21,4 +21,7 @@ from .download_utils import *
 from .prettyloss import *
 from . import nnets
 from . import model_metrics
-from . import xai
\ No newline at end of file
+from . import xai
+from . import dataset_analysis as dataAnalyzer
+from . import callbacks as callbacks
+from . import dataloaders as dataloader
\ No newline at end of file
