@@ -27,16 +27,28 @@ import numpy as np
 import matplotlib.pyplot as plt
 from  . import networks 
 from  . import dataset_prepare as dataprc 
-import plotly.graph_objects as go
-from plotly.subplots import make_subplots
+try:
+	import plotly.graph_objects as go
+	from plotly.subplots import make_subplots
+except:
+	print('[WARNING]: Plotly Not Installed. Plot Training Metrics wont work PLease install the library.')
 import pandas as pd
 import seaborn as sns
 from sklearn.metrics import accuracy_score,classification_report,cohen_kappa_score
 from .model_metrics import *
 from .xai import *
 from .uFuncs import *
+from albumentations import Compose
+import albumentations.augmentations.transforms as augmentations
+from .dataloaders import *
 
-	
+physical_devices = tf.config.list_physical_devices('GPU')
+if len(physical_devices)>1:
+	MULTI_GPU_MODE= True
+	print('[INFO]: Medicalai activated with MultiGPU Mode')
+else:
+	MULTI_GPU_MODE= False
+GPU_to_Use = 'all'	
 	
 def create_model_output_folder(outputName):
 	"""
@@ -70,10 +82,10 @@ def check_model_exists(outputName):
 
 	"""	
 	if not os.path.exists(outputName+'_arch.json'):
-		print("\n\n\n \tModel Doesnt Exist \n\n\n")
+		print("\n\n\n \t[INFO]: Model Doesnt Exist \n\n\n")
 		return False
 	else:
-		print("\n\n\n \tUsing Model: {:}.json \n\n\n".format(outputName))
+		print("\n\n\n \t[INFO]: Using Model: {:}.json \n\n\n".format(outputName))
 		return True
 		
 def save_model_and_weights(model,outputName):
@@ -132,7 +144,7 @@ def load_model_and_weights(modelName, summary = False):
 	model.load_weights(modelName+'_wgts.h5')
 	return model
 
-def modelManager(modelName,x_train, OUTPUT_CLASSES, RETRAIN_MODEL, AI_NAME= 'tinyMedNet', convLayers=None):
+def modelManager(modelName,x_train, OUTPUT_CLASSES, RETRAIN_MODEL, AI_NAME= 'tinyMedNet', **kwargs):
 	"""
 	Model manager is used to build new model for given networks/AI or reload existing AI model. 
 	This function can be used to retrain existing models or create new models.
@@ -163,12 +175,13 @@ def modelManager(modelName,x_train, OUTPUT_CLASSES, RETRAIN_MODEL, AI_NAME= 'tin
 		if check_model_exists(modelName):
 			model = load_model_and_weights(modelName = modelName)
 		else:
+			create_model_output_folder(modelName)
 			nw = networks.get(AI_NAME)
-			model = nw(x_train.shape[1:],OUTPUT_CLASSES, convLayers)
+			model = nw(x_train.shape[1:],OUTPUT_CLASSES, **kwargs)
 			
 	else:
 		nw = networks.get(AI_NAME)
-		model = nw(x_train.shape[1:],OUTPUT_CLASSES, convLayers)
+		model = nw(x_train.shape[1:],OUTPUT_CLASSES, **kwargs)
 	
 	return model
 	
@@ -186,15 +199,12 @@ def show_model_details(model):
 
 def train(	model, x_train, 
 			batch_size=1,epochs=1, 
-			learning_rate=0.001, callbacks=None,
+			callbacks=None,
 			class_weights = None, 
 			saveBestModel = False, bestModelCond = None, 
 			validation_data = None, TRAIN_STEPS = None, TEST_STEPS = None, 
-			loss = 'sparse_categorical_crossentropy', metrics = ['accuracy'], verbose=None, y_train=None, 
+			verbose=None, y_train=None, workers = 1 
 		  ):
-	model.compile(optimizer=tf.keras.optimizers.Adam(lr=learning_rate),
-				  loss=loss,
-				  metrics=metrics)
 	if callbacks is not None:
 		if ('tensorboard'in callbacks):
 			logdir=os.path.join('log')
@@ -204,7 +214,7 @@ def train(	model, x_train,
 
 	if saveBestModel is not None and saveBestModel==True:
 		if bestModelCond is not None and bestModelCond.lower() != 'default':
-			print('\n\tENGINE INITIALIZED WITH "SAVING BEST MODEL" and Early Stopping MODE\n')
+			print('\n[INFO]: ENGINE INITIALIZED WITH "SAVING BEST MODEL" and Early Stopping MODE\n')
 			earlystop_callback = tf.keras.callbacks.EarlyStopping(
 			monitor=bestModelCond['monitor'], 
 			min_delta=bestModelCond['min_delta'],
@@ -221,7 +231,8 @@ def train(	model, x_train,
 				epochs=epochs,
 				validation_data=validation_data,
 				callbacks=callbacks,
-				class_weight = class_weights
+				class_weight = class_weights,
+				workers =workers
 				)
 	else:
 		result = model.fit(x_train,
@@ -230,7 +241,8 @@ def train(	model, x_train,
 				validation_data=validation_data,
 				callbacks=callbacks,
 				verbose = verbose,
-				class_weight = class_weights
+				class_weight = class_weights,
+				workers = workers
 				)		
 	return result.history
 
@@ -346,10 +358,12 @@ class INFERENCE_ENGINE(object):
 		if modelName is not None:
 			self.model = load_model_and_weights(modelName = modelName)
 		self.labelNames = classNames
-		try:
-			self.preprocessor_from_meta()
-		except:
-			print('Meta File Not Found - Not Initializing Preprocessor from Meta')
+		print('[INFO]: Initialized -',self.__class__.__name__)
+		if self.__class__.__name__ == 'INFERENCE_ENGINE':
+			try:
+				self.preprocessor_from_meta()
+			except:
+				print('[INFO]: Meta File Not Found - Not Initializing Preprocessor from Meta')
 	
 	def load_model_and_weights(self, modelName, summary=False):
 		"""
@@ -487,7 +501,7 @@ class INFERENCE_ENGINE(object):
 			self.labelNames = self.preProcessor.labels
 
 	#@timeit
-	def predict(self, input):
+	def predict(self, input, verbose=1, safe=False  , workers= 1):
 		"""
 		Peform prediction on Input. Input can be Numpy Array or Image or Data Generator (in case of Test/Validation). 
 
@@ -514,22 +528,38 @@ class INFERENCE_ENGINE(object):
 		# Returns
 			Numpy.Array: of Predictions. Shape of Output [Number of Inputs, Number of Output Classes in Model]
 		"""
-		if self.preProcessor is not None:
-			input = self.preProcessor.processImage(input)
-			return self.model.predict(input)
+		
+		if hasattr(self, 'workers'):
+			workers = self.workers  
 		else:
-			if self.labelNames is None:
-				if hasattr(input, 'labelNames'):
-					self.labelNames = input.labelNames if self.labelNames is None else self.labelNames
-			if isinstance(input,np.ndarray):
-				return self.model.predict(input)
-			elif hasattr(input, 'data'):
-				return self.model.predict(input.data)
-			elif hasattr(input, 'generator'):
-				#self.labelNames =	dataprc.safe_labelmap_converter(input.labelMap) if labelNames is None else labelNames
-				return self.model.predict(input.generator)
+			workers = workers
+		if safe:
+			if isinstance(input,ImageSequenceFromDF):
+				return self.model.predict(input,  steps =(input.n/input.batch_size), verbose=1, workers=workers)
+			elif hasattr(input, 'generator') and hasattr(input, 'STEP_SIZE'):
+				return self.model.predict(input.generator, steps=input.STEP_SIZE, verbose=1, workers=workers)
+			elif hasattr(input, 'image_data_generator'):
+				return self.model.predict(input,  steps =(input.n/input.batch_size), verbose=1, workers=workers)
+		else:			
+			if hasattr(input, 'generator') and hasattr(input, 'STEP_SIZE'):
+				return self.model.predict(input.generator, verbose=1, workers=workers)
+			elif hasattr(input, 'image_data_generator') or isinstance(input,ImageSequenceFromDF):
+				#print('\n\nUsing INPUT For Predict:{}\n\n'.format(type(input)))
+				return self.model.predict(input,  verbose=1, workers=workers)
+			elif hasattr(input, 'data') and not isinstance(input,np.ndarray):
+				return self.model.predict(input.data, verbose=verbose, workers=workers)
 			else:
-				return self.model.predict(input)
+				if self.preProcessor is not None:
+					input = self.preProcessor.processImage(input)
+					return self.model.predict(input, verbose=verbose, workers=workers)
+				else:
+					if self.labelNames is None:
+						if hasattr(input, 'labelNames'):
+							self.labelNames = input.labelNames if self.labelNames is None else self.labelNames
+					if isinstance(input,np.ndarray):
+						return self.model.predict(input, verbose=verbose, workers=workers)
+					else:
+						return self.model.predict(input, verbose=verbose, workers=workers)
 
 	#@timeit
 	def predict_pipeline(self, input):
@@ -563,6 +593,13 @@ class INFERENCE_ENGINE(object):
 		img = self.preProcessor.processImage(input)
 		return self.model.predict(img)
 
+	def predict_with_labels(self, input,top_preds=4, retType = 'tuple'):
+		"""
+		Returns Decodes predictions with label/class names with output probabilites
+		"""
+		pred = self.predict(input)
+		return self.decode_predictions( pred, top_preds= top_preds, retType = retType)
+
 	def decode_predictions(self, pred, top_preds=4, retType = 'tuple'):
 		"""
 		Returns Decodes predictions with label/class names with output probabilites. 
@@ -656,7 +693,7 @@ class INFERENCE_ENGINE(object):
 		"""	
 		return self.model.summary()
 
-	def generate_evaluation_report(self, testSet = None, predictions = None, printStat = False,returnPlot = False, showPlot= False, pdfName =None, **kwargs):
+	def generate_evaluation_report(self, testSet = None, predictions = None, printStat = True,returnPlot = False, showPlot= False, pdfName =None, **kwargs):
 		"""
 		Generate a comprehensive PDF report with model sensitivity, specificity, accuracy, confidence intervals,
 		ROC Curve Plot, Precision Recall Curve Plot, and Confusion Matrix Plot for each class.
@@ -710,8 +747,9 @@ class INFERENCE_ENGINE(object):
 			predictions = self.predict(testSet)
 		if pdfName is None:
 			pdfName = self.modelName
+		
 		_get_metrics_evals(testSet, predictions, printStat = printStat,returnPlot = returnPlot, showPlot= showPlot,modelName=self.modelName, pdfName = pdfName, **kwargs)
-		print('Report Generated at Path:\n\t', os.path.abspath(pdfName)+'_report.pdf')
+		print('[INFO]: Report Generated at Path:\n\t', os.path.abspath(pdfName)+'_report.pdf')
 
 	def explain(self,input, predictions=None, layer_to_explain='CNN3', classNames = None, selectedClasses=None, expectedClass = None, showPlot=False):
 		"""
@@ -748,10 +786,14 @@ class INFERENCE_ENGINE(object):
 
 		"""
 		if predictions is None:
-			predictions = self.predict(input)#.model.predict(input)
-			
+			try:
+				img = self.preProcessor.processImage(input)
+				predictions = self.predict(img)#.model.predict(input)
+			except:
+				predictions = self.predict(input)
+				
 		if classNames is None and self.labelNames is None:
-			print('Error: No Labels Provides')
+			print('[ERROR]: No Labels Provides')
 		elif classNames is not None:
 			labels = classNames
 		elif self.labelNames is not None:
@@ -782,8 +824,12 @@ class TRAIN_ENGINE(INFERENCE_ENGINE):
 	"""
 	def __init__(self, modelName=None):
 		super().__init__(modelName)
-
-	def train_and_save_model(self,AI_NAME, MODEL_SAVE_NAME, trainSet, testSet, OUTPUT_CLASSES, RETRAIN_MODEL, BATCH_SIZE, EPOCHS, LEARNING_RATE, convLayers=None,SAVE_BEST_MODEL=False, BEST_MODEL_COND=None, callbacks=None, loss = 'sparse_categorical_crossentropy', metrics = ['accuracy'], showModel = False,CLASS_WEIGHTS=None):
+		
+	def train_and_save_model(self,AI_NAME, MODEL_SAVE_NAME, trainSet, testSet, OUTPUT_CLASSES, RETRAIN_MODEL,  EPOCHS, 
+							BATCH_SIZE=32, LEARNING_RATE=0.0001, SAVE_BEST_MODEL=False, BEST_MODEL_COND=None, 
+							callbacks=None, loss = 'sparse_categorical_crossentropy', optimizer=tf.keras.optimizers.Adam(lr=0.0001),
+							metrics = ['accuracy'], showModel = False, workers = 1,
+							CLASS_WEIGHTS=None, **kwargs):
 		""""
 		Main function that trains and saves a model. This automatically builds new model for given networks/AI or reload existing AI model. 
 		This function can be used to retrain existing models or create new models.
@@ -858,34 +904,68 @@ class TRAIN_ENGINE(INFERENCE_ENGINE):
 			None: On successful completion saves the trained model.
 
 		"""
+		self.workers = workers
 		self.testSet = testSet
 		self.modelName = MODEL_SAVE_NAME
 		self.test_predictions = None
+		global MULTI_GPU_MODE, GPU_to_Use
 		if hasattr(trainSet, 'data'):
 			self.labelNames = trainSet.labelNames
-			self.model = modelManager(AI_NAME= AI_NAME, convLayers= convLayers, modelName = MODEL_SAVE_NAME, x_train = trainSet.data, OUTPUT_CLASSES = OUTPUT_CLASSES, RETRAIN_MODEL= RETRAIN_MODEL)
+			if MULTI_GPU_MODE and GPU_to_Use.lower()=='all':
+				mirrored_strategy = tf.distribute.MirroredStrategy()
+				with mirrored_strategy.scope():
+					self.model = modelManager(AI_NAME= AI_NAME, modelName = MODEL_SAVE_NAME, x_train = trainSet.data, OUTPUT_CLASSES = OUTPUT_CLASSES, RETRAIN_MODEL= RETRAIN_MODEL, **kwargs)
+					self.model.compile(optimizer=optimizer,loss=loss,metrics=metrics)
+				#BATCH_SIZE *= mirrored_strategy.num_replicas_in_sync
+			else:
+					self.model = modelManager(AI_NAME= AI_NAME, modelName = MODEL_SAVE_NAME, x_train = trainSet.data, OUTPUT_CLASSES = OUTPUT_CLASSES, RETRAIN_MODEL= RETRAIN_MODEL, **kwargs)
+					self.model.compile(optimizer=optimizer,loss=loss,metrics=metrics)				
 			print(self.model.summary()) if showModel else None
-			
-			self.result = train(self.model, trainSet.data, y_train= trainSet.labels, batch_size=BATCH_SIZE, epochs=EPOCHS, learning_rate=LEARNING_RATE, 
+			print('[INFO]: BATCH_SIZE -',BATCH_SIZE)
+			self.result = train(self.model, trainSet.data, y_train= trainSet.labels, batch_size=BATCH_SIZE, epochs=EPOCHS,
 									validation_data=(testSet.data, testSet.labels), callbacks=callbacks, saveBestModel= SAVE_BEST_MODEL, 
-									bestModelCond = BEST_MODEL_COND, TRAIN_STEPS = None, TEST_STEPS = None, loss = loss, 
-									metrics = metrics,class_weights=CLASS_WEIGHTS)#['tensorboard'])
+									bestModelCond = BEST_MODEL_COND, TRAIN_STEPS = None, TEST_STEPS = None, workers = self.workers,
+									class_weights=CLASS_WEIGHTS)#['tensorboard'])
 			#self.model.evaluate(testSet.data, testSet.labels)
 			
 			dataprc.metaSaver(trainSet.labelMap, trainSet.labelNames,  normalize=trainSet.normalize,
 							  rescale =None,
 							  network_input_dim =trainSet.network_input_dim, samplingMethodName=trainSet.samplingMethodName, outputName= MODEL_SAVE_NAME)
 		else:
-			networkDim = np.zeros((1,)+trainSet.generator.image_shape)
+			from tensorflow.python.data.ops.dataset_ops import PrefetchDataset
+			if isinstance(trainSet.generator, PrefetchDataset):
+				for f,l in trainSet.generator.take(1):
+					inpSize = f.numpy().shape
+				networkDim = np.zeros((1,)+inpSize[1:])
+				networkInputSize = inpSize[1:]
+				rescaleValue = 1./255
+			else:
+				networkDim = np.zeros((1,)+trainSet.generator.image_shape)
+				networkInputSize = trainSet.generator.image_shape
+				try:
+					rescaleValue = trainSet.generator.image_data_generator.rescale
+				except:
+					rescaleValue = 1./255
+
 			self.labelNames = dataprc.safe_labelmap_converter(trainSet.labelMap)
-			self.model = modelManager(AI_NAME= AI_NAME, convLayers= convLayers, modelName = MODEL_SAVE_NAME, x_train = networkDim, OUTPUT_CLASSES = OUTPUT_CLASSES, RETRAIN_MODEL= RETRAIN_MODEL)
-			
+			if MULTI_GPU_MODE and GPU_to_Use.lower()=='all':
+				mirrored_strategy = tf.distribute.MirroredStrategy()
+				with mirrored_strategy.scope():
+					self.model = modelManager(AI_NAME= AI_NAME, modelName = MODEL_SAVE_NAME, x_train = networkDim, OUTPUT_CLASSES = OUTPUT_CLASSES, RETRAIN_MODEL= RETRAIN_MODEL, **kwargs)
+					self.model.compile(optimizer=optimizer,loss=loss,metrics=metrics)
+			else:
+					self.model = modelManager(AI_NAME= AI_NAME, modelName = MODEL_SAVE_NAME, x_train = networkDim, OUTPUT_CLASSES = OUTPUT_CLASSES, RETRAIN_MODEL= RETRAIN_MODEL, **kwargs)
+					self.model.compile(optimizer=optimizer,loss=loss,metrics=metrics)				
 			print(self.model.summary()) if showModel else None
-			self.result = train(self.model, trainSet.generator, batch_size=None, epochs=EPOCHS, learning_rate=LEARNING_RATE, validation_data=testSet.generator, callbacks=callbacks, saveBestModel= SAVE_BEST_MODEL, bestModelCond = BEST_MODEL_COND, TRAIN_STEPS = trainSet.STEP_SIZE, TEST_STEPS = testSet.STEP_SIZE, loss = loss, metrics =metrics, verbose=1,class_weights=CLASS_WEIGHTS)#['tensorboard'])
+			print('[INFO]: BATCH_SIZE -',BATCH_SIZE)
+			self.result = train(self.model, trainSet.generator, batch_size=BATCH_SIZE, epochs=EPOCHS, validation_data=testSet.generator, 
+					callbacks=callbacks, saveBestModel= SAVE_BEST_MODEL, bestModelCond = BEST_MODEL_COND, TRAIN_STEPS = trainSet.STEP_SIZE, 
+					TEST_STEPS = testSet.STEP_SIZE, verbose=1,class_weights=CLASS_WEIGHTS, workers = self.workers
+					)
 			#self.model.evaluate(testSet.generator,steps =	testSet.STEP_SIZE)
 			dataprc.metaSaver(trainSet.labelMap, self.labelNames, normalize= None,
-							 rescale = trainSet.generator.image_data_generator.rescale,
-							 network_input_dim =trainSet.generator.image_shape, samplingMethodName=None, outputName= MODEL_SAVE_NAME)
+							 rescale = rescaleValue,
+							 network_input_dim =networkInputSize, samplingMethodName=None, outputName= MODEL_SAVE_NAME)
 
 		save_model_and_weights(self.model, outputName= MODEL_SAVE_NAME)
 
