@@ -11,10 +11,12 @@
 #    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 #    See the License for the specific language governing permissions and
 #    limitations under the License.
-
 from __future__ import absolute_import
-from .nnets import resnet,covid_net,densenet,vgg16,mobilenet,mobilenetv2,xception,inceptionv3,inceptionResnet
 
+
+
+from .nnets import covid_net,densenet,vgg16,mobilenet,mobilenetv2,xception,inceptionv3,inceptionResnet
+from .nnets.resnet import resnet
 import os
 os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2' 
 import sys
@@ -24,6 +26,8 @@ from tensorflow.keras.layers import Dense, Dropout, Activation, Flatten
 from tensorflow.keras.layers import Conv2D, MaxPooling2D
 from tensorflow.keras.models import Sequential
 import tensorflow as tf
+
+ 
 class NetworkInit(object):
     """Base class for parameter Network initializers.
 
@@ -31,13 +35,13 @@ class NetworkInit(object):
     to initialize network/model parameters for numerous medical ai networks. It should be
     subclassed when implementing new types of network initializers.
     """
-    def __call__(self, inputSize, OutputSize, convLayers=None):
+    def __call__(self, inputSize, outputSize, **kwargs):
         """Makes :class:`NetworkInit` instances callable like a function, invoking
         their :meth:`call()` method.
         """
-        return self.call(inputSize, OutputSize, convLayers)
+        return self.call(inputSize, outputSize, **kwargs)
 
-    def call(self, inputSize, OutputSize, convLayers=None):
+    def call(self, inputSize, outputSize, **kwargs):
         """Sample should return model initialized with input and output Sizes.
         
         Parameters
@@ -45,7 +49,7 @@ class NetworkInit(object):
         inputSize : tuple or int.
             Integer or tuple specifying the input of network.
 
-        OutputSize : tuple or int.
+        outputSize : tuple or int.
             Integer or tuple specifying the output classes of network.
 
         Returns
@@ -63,7 +67,12 @@ class tinyMedNet(NetworkInit):
     """tinyMedNet is a classification network that consumes very less resources and can be trained even on CPUs. This network can be used to demonstrate the framework working.
        Additionally this acts a starting point for example/tutorial for getting started to know the Medical AI library.
     """
-    def call(self, inputSize, OutputSize, convLayers=None):
+    def call(self, inputSize, outputSize, **kwargs):
+        try:
+            finalActivation = kwargs["finalActivation"]
+            print('[INFO]: {} Final Activation Set to {}'.format(self.__class__.__name__,finalActivation))
+        except:
+            finalActivation = 'softmax'
         try:
             model = Sequential([
                 Conv2D(64, kernel_size=(5, 5), strides=(1, 1),activation='relu', padding = 'valid',input_shape=inputSize, name='CNN1'),
@@ -75,7 +84,7 @@ class tinyMedNet(NetworkInit):
                 Flatten(),
                 Dense(384, activation='relu', name='FC1'),
                 Dense(192, activation='relu', name='FC2'),
-                Dense(OutputSize, activation='softmax', name='FC3')
+                Dense(outputSize, activation=finalActivation, name='FC3')
             ])
         except ValueError:
             model = Sequential([
@@ -88,16 +97,28 @@ class tinyMedNet(NetworkInit):
                 Flatten(),
                 Dense(384, activation='relu', name='FC1'),
                 Dense(192, activation='relu', name='FC2'),
-                Dense(OutputSize, activation='softmax', name='FC3')
+                Dense(outputSize, activation='softmax', name='FC3')
             ])            
         return model
 
 class tinyMedNet_v2(NetworkInit):
     """tinyMedNet_v2 allows users to configure the number of Conv/CNN layers.
+       Should Pass `convLayers` when initializing.
        tinyMedNet_v2 is a classification network that consumes very less resources and can be trained even on CPUs. This network can be used to demonstrate the framework working.
        Additionally this acts a starting point for example/tutorial for getting started to know the Medical AI library.
     """
-    def call(self, inputSize, OutputSize, convLayers=2):
+    def call(self, inputSize, outputSize, **kwargs):
+        try:
+            convLayers = kwargs["convLayers"]
+        except:
+            print('convLayers Not Passed in Network Parameters. Pass the parameter using **ai_params')
+            print('Using Default 3 convLayers')
+            convLayers = 3
+        try:
+            finalActivation = kwargs["finalActivation"]
+            print('[INFO]: {} Final Activation Set to {}'.format(self.__class__.__name__,finalActivation))
+        except:
+            finalActivation = 'softmax'
         try:
             model = Sequential()
             model.add(Conv2D(64, kernel_size=(3, 3), strides=(1, 1),activation='relu', padding = 'valid',input_shape=inputSize, name='CNN1'))
@@ -108,7 +129,7 @@ class tinyMedNet_v2(NetworkInit):
             model.add(Flatten())
             model.add(Dense(384, activation='relu', name='FC1'))
             model.add(Dense(192, activation='relu', name='FC2'))
-            model.add(Dense(OutputSize, activation='softmax', name='FC3'))
+            model.add(Dense(outputSize, activation=finalActivation, name='FC3'))
             return model
         except ValueError:
             print(20*'-')
@@ -119,8 +140,21 @@ class tinyMedNet_v2(NetworkInit):
         
 class tinyMedNet_v3(NetworkInit):
     """tinyMedNet_v3 has 3 FC layers with Dropout and Configurable number of Conv/CNN Layers.
+       Should Pass `convLayers` when initializing.
     """
-    def call(self, inputSize, OutputSize, convLayers=2):
+    def call(self, inputSize, outputSize, **kwargs):
+        try:
+            convLayers = kwargs["convLayers"]
+        except:
+            print('convLayers Not Passed in Network Parameters. Pass the parameter using **ai_params')
+            print('Using Default 3 convLayers')
+            convLayers = 3
+        try:
+            finalActivation = kwargs["finalActivation"]
+            print('[INFO]: {} Final Activation Set to {}'.format(self.__class__.__name__,finalActivation))
+        except:
+            finalActivation = 'softmax'
+
         try:
             model = Sequential()
             model.add(Conv2D(64, kernel_size=(3, 3), strides=(1, 1),activation='relu', padding = 'valid',input_shape=inputSize, name='CNN1'))
@@ -136,7 +170,7 @@ class tinyMedNet_v3(NetworkInit):
             model.add(Dropout(rate=0.5))
             model.add(Dense(192, activation='relu', name='FC3'))
             model.add(Dropout(rate=0.5))
-            model.add(Dense(OutputSize, activation='softmax', name='FC4'))
+            model.add(Dense(outputSize, activation=finalActivation, name='FC4'))
             return model
         except ValueError as err:
             print(err)
@@ -146,33 +180,42 @@ class tinyMedNet_v3(NetworkInit):
             print(20*'-')
             sys.exit(1)            
         
-class resNet20(NetworkInit):
-    """resnet20
+class resNet50(NetworkInit):
+    """RESNET50
     """
-    def call(self, inputSize, OutputSize, convLayers=None):
-        img_input = tf.keras.layers.Input(shape=inputSize)
-        return resnet.resnet20(img_input=img_input,classes=OutputSize)
+    def call(self, inputSize, outputSize, **kwargs):
+        return resnet(img_input=inputSize,classes=outputSize, name='ResNet50', **kwargs)
 
-class resNet32(NetworkInit):
-    """resnet32
+class resNet50V2(NetworkInit):
+    """RESNET50V2
     """
-    def call(self, inputSize, OutputSize, convLayers=None):
-        img_input = tf.keras.layers.Input(shape=inputSize)
-        return resnet.resnet32(img_input=img_input,classes=OutputSize)
+    def call(self, inputSize, outputSize, **kwargs):
+        return resnet(img_input=inputSize,classes=outputSize, name='ResNet50V2', **kwargs)
 
-class resNet56(NetworkInit):
-    """RESNET56
+class resNet101(NetworkInit):
+    """RESNET101
     """
-    def call(self, inputSize, OutputSize, convLayers=None):
-        img_input = tf.keras.layers.Input(shape=inputSize)
-        return resnet.resnet56(img_input=img_input,classes=OutputSize)
+    def call(self, inputSize, outputSize, **kwargs):
+        return resnet(img_input=inputSize,classes=outputSize, name='ResNet101', **kwargs)
 
-class resNet110(NetworkInit):
-    """resnet110
+class resNet101V2(NetworkInit):
+    """RESNET101V2
     """
-    def call(self, inputSize, OutputSize, convLayers=None):
-        img_input = tf.keras.layers.Input(shape=inputSize)
-        return resnet.resnet110(img_input=img_input,classes=OutputSize)
+    def call(self, inputSize, outputSize, **kwargs):
+        return resnet(img_input=inputSize,classes=outputSize, name='ResNet101V2', **kwargs)
+
+class resNet152V2(NetworkInit):
+    """RESNET152V2
+    """
+    def call(self, inputSize, outputSize, **kwargs):
+        return resnet(img_input=inputSize,classes=outputSize, name='ResNet152V2', **kwargs)
+
+class resNet152(NetworkInit):
+    """RESNET152
+    """
+    def call(self, inputSize, outputSize, **kwargs):
+        return resnet(img_input=inputSize,classes=outputSize, name='ResNet152', **kwargs)
+
 
 class megaNet(NetworkInit):
     """
@@ -180,8 +223,8 @@ class megaNet(NetworkInit):
     This is a tensorflow 2.0 network variant for COVID-Net described in Paper "COVID-Net: A Tailored Deep Convolutional Neural Network Design for Detection of COVID-19 Cases from Chest Radiography Images" by Linda Wang et al. 
     Reference: https://github.com/busyyang/COVID-19/
     """
-    def call(self, inputSize, OutputSize, convLayers=None):
-        return covid_net.COVIDNET_Keras(img_input=inputSize,classes=OutputSize)
+    def call(self, inputSize, outputSize, **kwargs):
+        return covid_net.COVIDNET_Keras(img_input=inputSize,classes=outputSize)
 
 class DenseNet121(NetworkInit):
     """
@@ -189,8 +232,9 @@ class DenseNet121(NetworkInit):
     inputSize: input image size tuple
     outputSize: Number of classes for prediction
     """
-    def call(self, inputSize, OutputSize, convLayers=None):
-        return densenet.DenseNet121_Model(img_input=inputSize,classes=OutputSize)
+    def call(self, inputSize, outputSize, **kwargs):
+        finalActivation = kwargs["finalActivation"] if 'finalActivation' in kwargs else "softmax"
+        return densenet.DenseNet121_Model(img_input=inputSize,classes=outputSize,finalActivation=finalActivation)
 
 class VGG16(NetworkInit):
     """
@@ -198,8 +242,8 @@ class VGG16(NetworkInit):
     inputSize: input image size tuple,default : (224,223,3)
     outputSize: Number of classes for prediction
     """
-    def call(self, inputSize, OutputSize, convLayers=None):
-        return vgg16.VGG16_Model(img_input=inputSize,classes=OutputSize)
+    def call(self, inputSize, outputSize, **kwargs):
+        return vgg16.VGG16_Model(img_input=inputSize,classes=outputSize)
 
 class MobileNet(NetworkInit):
     """
@@ -207,8 +251,8 @@ class MobileNet(NetworkInit):
     inputSize: input image size tuple,default : (224,223,3)
     outputSize: Number of classes for prediction
     """
-    def call(self, inputSize, OutputSize, convLayers=None):
-        return mobilenet.MobileNet(img_input=inputSize,classes=OutputSize)
+    def call(self, inputSize, outputSize, **kwargs):
+        return mobilenet.MobileNet(img_input=inputSize,classes=outputSize)
 
 class MobileNetV2(NetworkInit):
     """
@@ -216,8 +260,8 @@ class MobileNetV2(NetworkInit):
     inputSize: input image size tuple,default : (224,223,3)
     outputSize: Number of classes for prediction
     """
-    def call(self, inputSize, OutputSize, convLayers=None):
-        return mobilenetv2.MobileNetV2(img_input=inputSize,classes=OutputSize)
+    def call(self, inputSize, outputSize, **kwargs):
+        return mobilenetv2.MobileNetV2(img_input=inputSize,classes=outputSize)
 
 class Xception(NetworkInit):
     """
@@ -225,8 +269,8 @@ class Xception(NetworkInit):
     inputSize: input image size tuple,default : (224,223,3)
     outputSize: Number of classes for prediction
     """
-    def call(self, inputSize, OutputSize, convLayers=None):
-        return xception.Xception(img_input=inputSize,classes=OutputSize)
+    def call(self, inputSize, outputSize, **kwargs):
+        return xception.Xception(img_input=inputSize,classes=outputSize)
 
 class InceptionV3(NetworkInit):
     """
@@ -234,8 +278,8 @@ class InceptionV3(NetworkInit):
     inputSize: input image size tuple,default : (224,223,3)
     outputSize: Number of classes for prediction
     """
-    def call(self, inputSize, OutputSize, convLayers=None):
-        return inceptionv3.InceptionV3(img_input=inputSize,classes=OutputSize)
+    def call(self, inputSize, outputSize, **kwargs):
+        return inceptionv3.InceptionV3(img_input=inputSize,classes=outputSize)
 
 class InceptionResNetV2(NetworkInit):
     """
@@ -243,8 +287,8 @@ class InceptionResNetV2(NetworkInit):
     inputSize: input image size tuple,default : (224,223,3)
     outputSize: Number of classes for prediction
     """
-    def call(self, inputSize, OutputSize, convLayers=None):
-        return inceptionResnet.InceptionResNetV2_Model(img_input=inputSize,classes=OutputSize)
+    def call(self, inputSize, outputSize, **kwargs):
+        return inceptionResnet.InceptionResNetV2_Model(img_input=inputSize,classes=outputSize)
 
 def get(networkInitialization):
     if networkInitialization.__class__.__name__ == 'str':
@@ -254,14 +298,18 @@ def get(networkInitialization):
             return tinyMedNet_v2()
         elif networkInitialization in ['tinyMedNet_v3', 'tiny_Medical_Network_v3']:
             return tinyMedNet_v3()
-        elif networkInitialization in ['resNet20', 'resnet20']:
-            return resNet20()
-        elif networkInitialization in ['resNet32', 'resnet32']:
-            return resNet32()
-        elif networkInitialization in ['resNet56', 'resnet56']:
-            return resNet56()
-        elif networkInitialization in ['resNet110', 'resnet110']:
-            return resNet110()
+        elif networkInitialization in ['resNet50', 'resnet50']:
+            return resNet50()
+        elif networkInitialization in ['resNet50V2', 'resnet50V2']:
+            return resNet50V2()
+        elif networkInitialization in ['resNet101', 'resnet101']:
+            return resNet101()
+        elif networkInitialization in ['resNet101V2', 'resnet101V2']:
+            return resNet101V2()
+        elif networkInitialization in ['resNet152', 'resnet152']:
+            return resNet152()
+        elif networkInitialization in ['resNet152V2', 'resnet152V2']:
+            return resNet152V2()
         elif networkInitialization in ['megaNet', 'meganet']:
             return megaNet()
         elif networkInitialization in ['densenet','DenseNet','DenseNet121']:
@@ -276,7 +324,7 @@ def get(networkInitialization):
             return Xception()
         elif networkInitialization in ['inception','InceptionV3','inceptionv3']:
             return InceptionV3()
-        elif networkInitialization in ['inceptionresnet','InceptionResNet','InceptionReset','InceptionResNetV2']:
+        elif networkInitialization in ['inceptionresnet','InceptionResNet','InceptionResNetV2']:
             return InceptionResNetV2()
         raise ValueError('Unknown network Initialization name: {}.'.format(networkInitialization))
 
